@@ -109,7 +109,7 @@ static ARGPARSE_OPTS opts[] =
   };
 
 
-/* Print usage information and and provide strings for help. */
+/* Print usage information and provide strings for help. */
 static const char *
 my_strusage( int level )
 {
@@ -280,7 +280,7 @@ query_swdb (estream_t out, const char *name, const char *current_version)
   gnupg_isotime_t value_date = {0};
   char *value_size = NULL;
   char *value_sha2 = NULL;
-  unsigned long value_size_ul;
+  unsigned long value_size_ul = 0;
   int status, i;
 
 
@@ -470,6 +470,7 @@ main (int argc, char **argv)
   /* Make sure that our subsystems are ready.  */
   i18n_init();
   init_common_subsystems (&argc, &argv);
+  gc_components_init ();
 
   /* Parse the command line. */
   pargs.argc  = &argc;
@@ -588,6 +589,18 @@ main (int argc, char **argv)
 	  es_putc ('\n', es_stderr);
 	  exit (2);
 	}
+      else if (!strcmp (fname, "all"))
+        {
+          if (cmd == aLaunch)
+            {
+              if (gc_component_launch (-1))
+                exit (1);
+            }
+          else
+            {
+              gc_component_kill (-1);
+            }
+        }
       else
         {
           /* Launch/Kill a given component.  */
@@ -616,7 +629,7 @@ main (int argc, char **argv)
       break;
 
     case aReload:
-      if (!fname)
+      if (!fname || !strcmp (fname, "all"))
 	{
           /* Reload all.  */
           gc_component_reload (-1);
